/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif

#ifdef HAVE_GETOPT_H
#define _GNU_SOURCE
#include <getopt.h>
#endif

#include <gtk/gtk.h>

#include <libxfce4util/libxfce4util.h>
#include <libxfcegui4/libxfcegui4.h>

#ifdef HAVE_LIBSTARTUP_NOTIFICATION
#define SN_API_NOT_YET_FROZEN 1
#include <libsn/sn.h>
#endif

#include <exo/exo-xsession-client.h>

#include "gtkxine.h"

#include "main.h"
#include "mainwin.h"
#include "playlist.h"
#include <xfmedia/xfmedia-settings.h>
#include "xfmedia-internal.h"
#include "remote.h"
#include "mediamarks.h"

static void mainwin_destroy_cb(GtkWidget *w, gpointer user_data);
static void vwin_destroy_cb(GtkWidget *w, gpointer user_data);

static XfMediaMainwin *g_mwin = NULL;
static pid_t main_pid = 0;
static gboolean quitting = FALSE;
static guint mwin_dest_id = 0, vwin_dest_id = 0;

void
xfmedia_quit(XfMediaMainwin *mwin, XfMediaQuitMode quit_mode)
{
    GtkWidget *main_window, *video_window;
    if(!mwin)
        exit(0);
    
    xfmedia_mediamarks_cleanup();
    xfmedia_remote_cleanup(mwin);
    xfmedia_plugins_unload();
    
    if(xfmedia_settings_get_bool("/xfmedia/playlist/save_on_exit_always")
            || mwin->plist->save_on_exit)
    {
        xfmedia_settings_save_playlist(mwin->plist->file_list);
    }
    xfmedia_settings_cleanup();
    
    /* save pointers */
    main_window = mwin->window;
    video_window = mwin->video_window;
    
    xfmedia_mainwin_destroy(mwin);
    
    switch(quit_mode) {
        case XFMEDIA_QUIT_MODE_EXT:
            g_signal_handler_disconnect(G_OBJECT(main_window), mwin_dest_id);
            g_signal_handler_disconnect(G_OBJECT(video_window), vwin_dest_id);
            gtk_widget_destroy(video_window);
            gtk_widget_destroy(main_window);
            break;
        
        case XFMEDIA_QUIT_MODE_MAINWIN_DESTROY:
            g_signal_handler_disconnect(G_OBJECT(video_window), vwin_dest_id);
            gtk_widget_destroy(video_window);
            break;
        
        case XFMEDIA_QUIT_MODE_VWIN_DESTROY:
            g_signal_handler_disconnect(G_OBJECT(main_window), mwin_dest_id);
            gtk_widget_destroy(main_window);
            break;
    }
    
    if(gtk_main_level())
        gtk_main_quit();
    
    exit(0);
}

static void
mainwin_destroy_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    DBG("entering");
    
    xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_MAINWIN_DESTROY);
}

static void
vwin_destroy_cb(GtkWidget *w, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    DBG("entering");
    
    xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_VWIN_DESTROY);
}

static void
sighandler_cb(int sig)
{
    if(quitting)
        return;
    
    quitting = TRUE;
    
    if(main_pid == getpid())
        xfmedia_quit(g_mwin, XFMEDIA_QUIT_MODE_EXT);
}

static void
session_save_yourself_cb(ExoXsessionClient *client, gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    
    xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_EXT);
}

static void
do_startup_notify(GtkWidget *window)
{
#ifdef HAVE_LIBSTARTUP_NOTIFICATION
    SnDisplay *sdpy;
    SnLauncheeContext *ctx;
    
    sdpy = sn_display_new(GDK_DISPLAY(),
            (SnDisplayErrorTrapPush)gdk_error_trap_push,
            (SnDisplayErrorTrapPop)gdk_error_trap_pop);
    ctx = sn_launchee_context_new_from_environment(sdpy,
            gdk_screen_get_number(gtk_widget_get_screen(window)));
    if(ctx) {
        sn_launchee_context_setup_window(ctx, GDK_WINDOW_XWINDOW(window->window));
        sn_launchee_context_complete(ctx);
        sn_launchee_context_unref(ctx);
    }
    sn_display_unref(sdpy);
#endif
}

static gboolean
initial_autoplay(gpointer data)
{
    XfMediaMainwin *mwin = data;
    GtkTreePath *path;
    
    gdk_threads_enter();
    path = gtk_tree_path_new_first();
    xfmedia_playlist_play_file_at_path(mwin, path);
    gtk_tree_path_free(path);
    gdk_threads_leave();
    
    return FALSE;
}

enum
{
    OPT_DISPLAY = 1000,
    OPT_SMCLIENTID,
    OPT_DEBUG,
    OPT_SESSIONID,
    OPT_ENQUEUE
};

static const gchar *short_opts = "Vhvfp";
static const struct option long_opts[] =
{
    /* stuff we just want out of the way */
    { "display", required_argument, NULL, OPT_DISPLAY },
    { "sm-client-id", required_argument, NULL, OPT_SMCLIENTID },
    /* stuff we care about */
    { "version", no_argument, NULL, 'V' },
    { "help", no_argument, NULL, 'h' },
    { "verbose", no_argument, NULL, 'v' },
    { "debug", no_argument, NULL, OPT_DEBUG },
    { "fullscreen", no_argument, NULL, 'f' },
    { "autoplay", no_argument, NULL, 'p' },
    { "session", required_argument, NULL, OPT_SESSIONID },
    { "enqueue", no_argument, NULL, OPT_ENQUEUE },
    { NULL, 0, NULL, 0 }
};

typedef struct
{
    gboolean is_remote_cmd;
    gint session_id;
    gboolean is_enqueue;
    
    gint verbose_level;
    gboolean use_fullscreen;
    gboolean use_autoplay;
} OptionData;

static gint
xfmedia_parse_options(int argc, char **argv, OptionData *odata)
{
    gint c = 0, option_index = 0;
    
    opterr = 0;
    for(;;) {
        c = getopt_long(argc, argv, short_opts, long_opts, &option_index);
        if(c == -1)
            break;
        
        switch(c) {
            case OPT_DISPLAY:
            case OPT_SMCLIENTID:
                break;
            
            case 'V':
                g_print("Xfmedia version %s, Copyright (c) Brian Tarricone, <bjt23@cornell.edu>\n", VERSION);
                g_print("Released under the terms of the GNU General Public License.\n");
                g_print("Compiled against xine-lib %s, using xine-lib %s.\n",
                        XINE_VERSION, xine_get_version_string());
                if((gint)LIBXFCE4UTIL_MICRO_VERSION == LIBXFCE4UTIL_MICRO_VERSION) {
                    g_print("Compiled against Xfce %d.%d.%d, using Xfce %s.\n",
                            LIBXFCE4UTIL_MAJOR_VERSION, LIBXFCE4UTIL_MINOR_VERSION,
                            (gint)LIBXFCE4UTIL_MICRO_VERSION, xfce_version_string());
                } else {
                    g_print("Compiled against Xfce %d.%d.%.1f, using Xfce %s.\n",
                            LIBXFCE4UTIL_MAJOR_VERSION, LIBXFCE4UTIL_MINOR_VERSION,
                            (gdouble)LIBXFCE4UTIL_MICRO_VERSION, xfce_version_string());
                }
                
                exit(0);
                break;
            
            case 'h':
                g_print("I should print out a help string here.\n");
                exit(0);
                break;
            
            case 'v':
                odata->verbose_level = XINE_VERBOSITY_LOG;
                break;
            
            case OPT_DEBUG:
                odata->verbose_level = XINE_VERBOSITY_DEBUG;
                break;
            
            case 'f':
                odata->use_fullscreen = TRUE;
                break;
            
            case 'p':
                odata->use_autoplay = TRUE;
                break;
            
            case OPT_SESSIONID:
                odata->is_remote_cmd = TRUE;
                odata->session_id = atoi(optarg);
                break;
            
            case OPT_ENQUEUE:
                odata->is_remote_cmd = TRUE;
                odata->is_enqueue = TRUE;
                break;
            
            default:
                break;
        }
    }
    
    return optind;
}

int
main(int argc, char **argv)
{
    GtkWidget *gtx, *video_window;
    XfMediaMainwin *mwin;
    gint i, fd;
    gboolean defer = FALSE, do_autoplay = FALSE;
    unsigned int seed;
    GtkTreePath *path;
    gint x, y, ppos, next_arg;
    ExoXsessionClient *session_client;
    OptionData odata;
    gchar *session_argv[1];
    
    main_pid = getpid();
    
    signal(SIGINT, sighandler_cb);
    signal(SIGHUP, sighandler_cb);
    signal(SIGQUIT, sighandler_cb);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGUSR1, SIG_IGN);
    signal(SIGPIPE, SIG_IGN);
    
    memset(&odata, 0, sizeof(odata));
    odata.verbose_level = XINE_VERBOSITY_NONE;
    odata.session_id = -1;
    next_arg = xfmedia_parse_options(argc, argv, &odata);
    
    /* handle remote commands */
    if(odata.is_remote_cmd) {
        if(odata.is_enqueue && next_arg < argc) {
            while(next_arg < argc) {
                if(!xfmedia_remote_send_command(odata.session_id,
                        "add_file", argv[next_arg++], -1))
                {
                    return 1;
                }
            }
        } else {
            /* implement other stuff here */
        }
        
        return 0;
    }
    
    fd = open("/dev/random", O_RDONLY);
    if(fd >= 0) {
        read(fd, &seed, sizeof(unsigned int));
        close(fd);
    } else {
        fd = open("/dev/urandom", O_RDONLY);
        if(fd >= 0) {
            read(fd, &seed, sizeof(unsigned int));
            close(fd);
        } else
            seed = time(NULL) ^ (main_pid + (main_pid << 15));
    }
#if HAVE_SRANDOM
    srandom(seed);
#else
    srand(seed);
#endif
    
    g_thread_init(NULL);
    gdk_threads_init();
    gtk_init(&argc, &argv);
    
    xfce_textdomain(GETTEXT_PACKAGE, LOCALEDIR, "UTF-8");
    
    xfmedia_settings_init();
    xfmedia_init_icon_theme();
    xfmedia_stock_icons_init();
    xfmedia_mediamarks_init();
    
    gdk_threads_enter();
    
    video_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_window_set_title(GTK_WINDOW(video_window), "Xfmedia");
    gtk_container_set_border_width(GTK_CONTAINER(video_window), 0);
    vwin_dest_id = g_signal_connect(G_OBJECT(video_window), "destroy",
            G_CALLBACK(vwin_destroy_cb), mwin);
    gtk_widget_realize(video_window);
    
    gtx = gtk_xine_new("auto", "auto");
    gtk_widget_show(gtx);
    gtk_container_add(GTK_CONTAINER(video_window), gtx);
    DBG("about to realize...");
    gtk_widget_realize(gtx);
    DBG("    realize done");
    
    gtk_xine_engine_set_param(GTK_XINE(gtx),
            XINE_ENGINE_PARAM_VERBOSITY, odata.verbose_level);
    
    mwin = g_mwin = xfmedia_mainwin_new(GTK_XINE(gtx), GTK_WINDOW(video_window));
    mwin_dest_id = g_signal_connect(G_OBJECT(mwin->window), "destroy",
            G_CALLBACK(mainwin_destroy_cb), mwin);
    mwin->plist->is_dirty = FALSE;
    if(argc-next_arg > 11)
        defer = TRUE;
    for(i = next_arg; i < argc; i++) {
        if(xfmedia_playlist_is_playlist_file(argv[i]))
            xfmedia_playlist_load(argv[i], mwin, FALSE);
        else {
            xfmedia_mainwin_add_file(mwin, argv[i], -1, defer);
            mwin->plist->is_dirty = TRUE;
            mwin->plist->save_on_exit = FALSE;
        }
        do_autoplay = TRUE;
    }
    
    if(argc <= 1)
        xfmedia_settings_load_playlist(mwin);
    
    /* set up the shuffle bitmap.  this is somewhat redundant, depending, but
     * i want to make sure it gets done properly. (FIXME: startup perf) */
    xfmedia_playlist_shuffle_init_entries(mwin->plist,
            gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mwin->plist->file_list), NULL));
    
    x = xfmedia_settings_get_int("/xfmedia/general/pos_x");
    y = xfmedia_settings_get_int("/xfmedia/general/pos_y");
    if(x != -1 && y != -1)
        gtk_window_move(GTK_WINDOW(mwin->window), x, y);
    gtk_widget_show_all(mwin->window);
    gtk_widget_set_size_request(mwin->window, -1, -1);
    
    do_startup_notify(mwin->window);
    
    session_argv[0] = argv[0];
    session_client = exo_xsession_client_new_with_group(mwin->window->window);
    exo_xsession_client_set_restart_command(session_client, session_argv, 1);
    g_signal_connect(G_OBJECT(session_client), "save-yourself",
            G_CALLBACK(session_save_yourself_cb), mwin);
    
    gtk_widget_realize(mwin->plist->treeview);
    ppos = xfmedia_settings_get_int("/xfmedia/playlist/playlist_position");
    if(gtk_tree_model_iter_n_children(GTK_TREE_MODEL(mwin->plist->file_list), NULL) > 0
             && ppos != -1)
    {
        path = gtk_tree_path_new_from_indices(ppos, -1);
        mwin->plist->cur_playing =
                gtk_tree_row_reference_new(GTK_TREE_MODEL(mwin->plist->file_list),
                        path);
        gtk_tree_view_scroll_to_cell(GTK_TREE_VIEW(mwin->plist->treeview), path,
                NULL, TRUE, 0.5, 0.0);
        gtk_tree_path_free(path);
    }
    
    if(xfmedia_settings_get_bool("/xfmedia/general/show_tray_icon"))
        gtk_widget_show(mwin->tray_icon);
    
    xfmedia_remote_init(mwin);
    xfmedia_plugins_load(mwin);
    
    xfmedia_playlist_treeview_scroll_cb(NULL, mwin);
    if(odata.use_fullscreen)
        gtk_xine_set_fullscreen(mwin->gtx, TRUE);
    if(do_autoplay || odata.use_autoplay)
        g_idle_add(initial_autoplay, mwin);
    
    gtk_main();
    
    gdk_threads_leave();
    
    /* xfmedia_quit() handles the cleanup */
    
    return 0;
}
