/*
 *  xfmedia - simple gtk2 media player based on xine
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif

#ifdef HAVE_SYS_UN_H
#include <sys/un.h>
#endif

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif

#if defined(HAVE_STDARG_H)
#include <stdarg.h>
#elif defined(HAVE_VARARGS_H)
#include <varargs.h>
#define va_start(v, l) va_start(v)
#else
#error Your system does not support variable-argument functions.
#endif

#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX 108
#endif

#include <libxfce4util/libxfce4util.h>

#include "remote.h"
#include "playlist.h"
#include "mainwin.h"
#include "main.h"
#include "xfmedia-internal.h"

#define SOCKET_FORMAT "%s/xfmedia_remote.%d.%d"

static gboolean
remote_socket_conn_io_watch(GIOChannel *ioch, GIOCondition cond,
        gpointer user_data)
{
    XfMediaMainwin *mwin = user_data;
    gchar *line = NULL;
    gsize len;
    GError *err = NULL;
    GIOStatus ret;
    gint send_retcode = -1;
    gboolean do_close = FALSE, do_quit = FALSE;
    
    if(cond != G_IO_IN)
        return FALSE;  /* connection error, remove the source */
    
    DBG("got new connection, reading line");
    
    ret = g_io_channel_read_line(ioch, &line, &len, NULL, &err);
    if(ret != G_IO_STATUS_NORMAL) {
        if(err) {
            g_warning("Error on remote control connection (%d): %s",
                    err->code, err->message);
            g_error_free(err);
        } else 
            g_warning("Connection error on remote control, closing.");
        return FALSE;
    }
    
    if(!len || !line)
        return TRUE;
    
    DBG("read line: '%s'", line);
    
    /* parse the commands! */
    if(!g_ascii_strncasecmp(line, "clear_playlist", 14)) {
        gtk_list_store_clear(mwin->plist->file_list);
        xfmedia_playlist_shuffle_init_entries(mwin->plist, 0);
        send_retcode = 1;
        mwin->plist->is_dirty = TRUE;
        mwin->plist->save_on_exit = TRUE;
    } else if(!g_ascii_strncasecmp(line, "load_playlist", 12)) {
        gchar *filename = NULL;
        
        send_retcode = 0;
        ret = g_io_channel_read_line(ioch, &filename, NULL, NULL, NULL);
        if(ret == G_IO_STATUS_NORMAL && filename) {
            filename[strlen(filename)-1] = 0;
            xfmedia_playlist_load(filename, mwin, FALSE);
            send_retcode = 1;
            mwin->plist->is_dirty = TRUE;
            mwin->plist->save_on_exit = TRUE;
        }
        
        if(filename)
            g_free(filename);
    } else if(!g_ascii_strncasecmp(line, "save_playlist", 12)) {
        gchar *filename = NULL;
        
        send_retcode = 0;
        ret = g_io_channel_read_line(ioch, &filename, NULL, NULL, NULL);
        if(ret == G_IO_STATUS_NORMAL && filename) {
            filename[strlen(filename)-1] = 0;
            xfmedia_playlist_save(mwin->plist->file_list, filename);
            send_retcode = 1;
        }
        
        if(filename)
            g_free(filename);
    } else if(!g_ascii_strncasecmp(line, "add_file", 8)) {
        gchar *filename = NULL, *index_str = NULL;
        gint index = -1;
        
        DBG("got add_file command");
        
        send_retcode = 0;
        ret = g_io_channel_read_line(ioch, &filename, NULL, NULL, NULL);
        if(ret == G_IO_STATUS_NORMAL && filename) {
            filename[strlen(filename)-1] = 0;
            DBG("read filename: '%s'", filename);
            ret = g_io_channel_read_line(ioch, &index_str, NULL, NULL, NULL);
            if(ret == G_IO_STATUS_NORMAL && index_str) {
                index_str[strlen(index_str)] = 0;
                index = atoi(index_str);
                DBG("adding file with index %d", index);
                xfmedia_mainwin_add_file(mwin, filename, index, TRUE);
                xfmedia_playlist_shuffle_add_entries(mwin->plist, index, 1);
                send_retcode = 1;
                mwin->plist->is_dirty = TRUE;
                mwin->plist->save_on_exit = TRUE;
            }
        }
        
        if(filename)
            g_free(filename);
        if(index_str)
            g_free(index_str);
    } else if(!g_ascii_strncasecmp(line, "remove_file", 11)) {
        gchar *index_str;
        gint index;
        
        send_retcode = 0;
        ret = g_io_channel_read_line(ioch, &index_str, NULL, NULL, NULL);
        if(ret == G_IO_STATUS_NORMAL && index_str) {
            index_str[strlen(index_str)] = 0;
            index = atoi(index_str);
            if(index >= 0) {
                xfmedia_playlist_remove_file(mwin->plist, NULL, index);
                xfmedia_playlist_renumber_file_list(mwin->plist->file_list,
                        index, -1);
                xfmedia_playlist_shuffle_remove_entries(mwin->plist, index, 1);
                send_retcode = 1;
                mwin->plist->is_dirty = TRUE;
                mwin->plist->save_on_exit = TRUE;
            }
        }
        
        if(index_str)
            g_free(index_str);
    } else if(!g_ascii_strncasecmp(line, "play", 4)) {
        gchar *index_str;
        gint index;
        
        send_retcode = 0;
        ret = g_io_channel_read_line(ioch, &index_str, NULL, NULL, NULL);
        if(ret == G_IO_STATUS_NORMAL && index_str) {
            GtkTreePath *path = NULL;
            
            index_str[strlen(index_str)] = 0;
            index = atoi(index_str);
            if(index < 0) {
                if(gtk_xine_get_status(mwin->gtx) == XINE_STATUS_PLAY
                        && gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED) != XINE_SPEED_NORMAL)
                {
                    gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
                    send_retcode = 1;
                } else if(mwin->plist->cur_playing)
                    path = gtk_tree_row_reference_get_path(mwin->plist->cur_playing);
                else
                    path = gtk_tree_path_new_first();
            } else
                path = gtk_tree_path_new_from_indices(index, -1);
            
            if(path) {
                if(xfmedia_playlist_play_file_at_path(mwin, path))
                    send_retcode = 1;
                else
                    send_retcode = 0;
                gtk_tree_path_free(path);
            }
        }
        
        if(index_str)
            g_free(index_str);
    } else if(!g_ascii_strncasecmp(line, "pause", 5)) {
        if(gtk_xine_get_status(mwin->gtx) == XINE_STATUS_PLAY) {
            if(gtk_xine_get_param(mwin->gtx, XINE_PARAM_SPEED) == XINE_SPEED_PAUSE)
                gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, XINE_SPEED_NORMAL);
            else
                gtk_xine_set_param(mwin->gtx, XINE_PARAM_SPEED, XINE_SPEED_PAUSE);
            send_retcode = 1;
        } else
            send_retcode = 0;
    } else if(!g_ascii_strncasecmp(line, "prev", 4)) {
        xfmedia_prev_cb(NULL, mwin);
        send_retcode = 1;
    } else if(!g_ascii_strncasecmp(line, "next", 4)) {
        xfmedia_next_cb(NULL, mwin);
        send_retcode = 1;
    } else if(!g_ascii_strncasecmp(line, "close", 5)) {
        do_close = TRUE;
        send_retcode = 1;
    } else if(!g_ascii_strncasecmp(line, "quit", 4)) {
        do_quit = TRUE;
        send_retcode = 1;
    }
    
    
    if(send_retcode == 0) {
        DBG("failed, sending zero retcode");
        g_io_channel_write_chars(ioch, "0\n", 2, NULL, NULL);
    } else if(send_retcode == 1) {
        DBG("succeeded, sending zero retcode");
        g_io_channel_write_chars(ioch, "1\n", 2, NULL, NULL);
    }
    g_io_channel_flush(ioch, NULL);
    DBG("retcode passed");
    
    if(line)
        g_free(line);
    
    if(do_quit)
        xfmedia_quit(mwin, XFMEDIA_QUIT_MODE_EXT);
    
    return ((do_close || do_quit) ? FALSE : TRUE);
}

static gboolean
remote_socket_io_watch(GIOChannel *ioch, GIOCondition cond, gpointer user_data)
{
    gint conn_sock;
    GIOChannel *conn_ioch;
    struct sockaddr_un conn_addr;
    socklen_t addrlen = sizeof(conn_addr);
    
    if(cond != G_IO_IN)
        return TRUE;

    conn_sock = accept(g_io_channel_unix_get_fd(ioch),
            (struct sockaddr *)&conn_addr, &addrlen);
    if(conn_sock >= 0) {
        conn_ioch = g_io_channel_unix_new(conn_sock);
        g_io_channel_set_close_on_unref(conn_ioch, TRUE);
        g_io_add_watch(conn_ioch, G_IO_IN|G_IO_HUP|G_IO_ERR|G_IO_NVAL,
                remote_socket_conn_io_watch, user_data);
        g_io_channel_unref(conn_ioch);
    }
    
    return TRUE;
}

static gboolean
remote_check_existing_socket_avail(const struct sockaddr_un *addr)
{
    gint sock;
    gboolean ret = FALSE;
    
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if(sock < 0)
        return FALSE;
    
    fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
    
    errno = 0;
    if(connect(sock, (const struct sockaddr *)&addr, sizeof(struct sockaddr_un))
            && errno != EINPROGRESS)
    {
        ret = TRUE;
    }
    
    if(sock >= 0)
        close(sock);
    
    return ret;
}

static gint
remote_get_new_session_id()
{
    gint id;
    struct sockaddr_un addr;
    
    addr.sun_family = AF_UNIX;
    for(id = 0; id < 1024; id++) {
        g_snprintf(addr.sun_path, PATH_MAX, SOCKET_FORMAT,
                g_get_tmp_dir(), (gint)getuid(), id);
        if(!g_file_test(addr.sun_path, G_FILE_TEST_EXISTS)
                || remote_check_existing_socket_avail(&addr))
        {
            unlink(addr.sun_path);
            return id;
        }
    }
    
    g_critical("Unable to find usable unix socket below 1024; disabling remote control.");
    
    return -1;
}

void
xfmedia_remote_init(XfMediaMainwin *mwin)
{
    struct sockaddr_un addr;
    GIOChannel *ioch;
    
    *addr.sun_path = 0;
    
    mwin->session_id = remote_get_new_session_id();
    if(mwin->session_id == -1)
        goto cleanuperr;
    
    mwin->session_sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if(mwin->session_sock < 0) {
        g_critical("Unable to allocate socket; disabling remote control.");
        goto cleanuperr;
    }
    
    addr.sun_family = AF_UNIX;
    g_snprintf(addr.sun_path, UNIX_PATH_MAX, SOCKET_FORMAT,
            g_get_tmp_dir(), getuid(), mwin->session_id);
    if(bind(mwin->session_sock, (const struct sockaddr *)&addr, sizeof(addr))) {
        g_critical("Unable to bind to socket; disabling remote control.");
        goto cleanuperr;
    }
    
    if(listen(mwin->session_sock, SOMAXCONN)) {
        g_critical("Unable to set socket to listen mode; disabling remote control.");
        goto cleanuperr;
    }
    
    ioch = g_io_channel_unix_new(mwin->session_sock);
    if(!ioch) {
        g_critical("Unable to wrap socket in GIOChannel; disabling remote control.");
        goto cleanuperr;
    }
    g_io_channel_set_flags(ioch, G_IO_FLAG_NONBLOCK, NULL);
    g_io_channel_set_close_on_unref(ioch, FALSE);
    mwin->session_sock_sourceid = g_io_add_watch(ioch, G_IO_IN,
            remote_socket_io_watch, mwin);
    g_io_channel_unref(ioch);
    
    /* success */
    return;
    
    cleanuperr:  /* failure */
    
    if(mwin->session_sock >= 0)
        close(mwin->session_sock);
    mwin->session_sock = -1;
    mwin->session_id = -1;
    if(strlen(addr.sun_path))
        unlink(addr.sun_path);
}

void
xfmedia_remote_cleanup(XfMediaMainwin *mwin)
{
    if(mwin->session_sock_sourceid) {
        g_source_remove(mwin->session_sock_sourceid);
        mwin->session_sock_sourceid = 0;
    }
    
    if(mwin->session_sock >= 0) {
        gchar sockfile[PATH_MAX];
        
        shutdown(mwin->session_sock, SHUT_RDWR);
        close(mwin->session_sock);
        
        g_snprintf(sockfile, UNIX_PATH_MAX, SOCKET_FORMAT,
                g_get_tmp_dir(), getuid(), mwin->session_id);
        unlink(sockfile);
    }
    
    mwin->session_id = mwin->session_sock = -1;
}


/*******************************
 * client interface
 *******************************/

static gint
xfmedia_remote_find_open_socket()
{
    gint sock, i;
    struct sockaddr_un addr;
        
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if(sock < 0)
        return -1;
    
    for(i = 0; i < 1024; i++) {
        addr.sun_family = AF_UNIX;
        g_snprintf(addr.sun_path, UNIX_PATH_MAX, SOCKET_FORMAT,
                g_get_tmp_dir(), getuid(), i);
        if(!connect(sock, (const struct sockaddr *)&addr, sizeof(addr)))
            break;
    }
    
    close(sock);
    
    return (i < 1025 ? i : -1);
}

gboolean
#if defined(HAVE_STDARG_H)
xfmedia_remote_send_command(gint session_id, const gchar *command, ...)
#elif defined(HAVE_VARARGS_H)
xfmedia_remote_send_command(gint session_id, const gchar *command, va_alist) va_dcl
#endif
{
    va_list ap;
    gchar *arg_s, fullcmd[2048], buf[16];
    gint arg_i, sock;
    gboolean cmd_ok = FALSE;
    struct sockaddr_un addr;
    
    if(session_id < 0) {
        session_id = xfmedia_remote_find_open_socket();
        if(session_id < 0)
            return FALSE;
    }
    
    va_start(ap, command);
    
    if(!g_ascii_strcasecmp(command, "clear_playlist")
            || !g_ascii_strcasecmp(command, "pause")
            || !g_ascii_strcasecmp(command, "stop")
            || !g_ascii_strcasecmp(command, "prev")
            || !g_ascii_strcasecmp(command, "next")
            || !g_ascii_strcasecmp(command, "close")
            || !g_ascii_strcasecmp(command, "quit"))
    {
        /* commands with no arguments */
        g_snprintf(fullcmd, 2048, "%s\n", command);
        cmd_ok = TRUE;
    } else if(!g_ascii_strcasecmp(command, "load_playlist")
            || !g_ascii_strcasecmp(command, "save_playlist"))
    {
        /* commands with a single string arg */
        arg_s = va_arg(ap, gchar *);
        if(arg_s) {
            g_snprintf(fullcmd, 2048, "%s\n%s\n", command, arg_s);
            cmd_ok = TRUE;
        }
    } else if(!g_ascii_strcasecmp(command, "add_file")) {
        /* command with a string arg followed by an int arg */
        arg_s = va_arg(ap, gchar *);
        arg_i = va_arg(ap, gint);
        if(arg_s) {
            g_snprintf(fullcmd, 2048, "%s\n%s\n%d\n", command, arg_s, arg_i);
            cmd_ok = TRUE;
        }
    } else if(!g_ascii_strcasecmp(command, "remove_file")
            || !g_ascii_strcasecmp(command, "play"))
    {
        /* commands with a single int arg */
        arg_i = va_arg(ap, gint);
        g_snprintf(fullcmd, 2048, "%s\n%d\n", command, arg_i);
        cmd_ok = TRUE;
    }
    
    va_end(ap);
    
    if(!cmd_ok)
        return FALSE;
    
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if(sock < 0)
        return FALSE;
    
    DBG("connecting...");
    
    addr.sun_family = AF_UNIX;
    g_snprintf(addr.sun_path, UNIX_PATH_MAX, SOCKET_FORMAT, g_get_tmp_dir(),
            getuid(), session_id);
    if(connect(sock, (const struct sockaddr *)&addr, sizeof(addr))) {
        close(sock);
        return FALSE;
    }
    DBG("connected, sending '%s'", fullcmd);
    
    if(send(sock, fullcmd, strlen(fullcmd), 0) != strlen(fullcmd)) {
        close(sock);
        return FALSE;
    }
    DBG("send successful, waiting for reply");
    
    if(recv(sock, buf, 16, 0) != 2) {
        close(sock);
        return FALSE;
    }
    DBG("got good reply");
    
    close(sock);
    
    return (buf[0] == '1' ? TRUE : FALSE);
}
