/*
 *  xfmusic4-cli - simple cmdline test implementation using libxfmusic4
 *
 *  Copyright (c) 2004 Brian Tarricone, <bjt23@cornell.edu>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Library General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <termios.h>
#include <unistd.h>
#include <fcntl.h>
#define _GNU_SOURCE
#include <getopt.h>

#include <glib.h>
#include <glib/gthread.h>

#include <libxfce4util/i18n.h>
#include <libxfce4util/debug.h>
#include <libxfmusic4/libxfmusic4.h>

#include "cli-misc.h"

/* stupid, but getting round() out of math.h is a pain */
#define round(x) ( ((gdouble)x-(gint)x) >= 0.5000 ? ((gint)x)+1 : (gint)x )

static const gchar *shortopts = "vqhVo:i:";
static const struct option longopts[] = {
    { "verbose", no_argument, 0, 'v' },
    { "quiet", no_argument, 0, 'q' },
    { "help", no_argument, 0, 'h' },
    { "version", no_argument, 0, 'V' },
    { "input", required_argument, 0, 'i' },
    { "output", required_argument, 0, 'o' }
};

static void
stream_ended_cb(XfMusic *xfmusic, gpointer user_data)
{
    GMainLoop *mloop = user_data;
    
    if(g_main_loop_is_running(mloop))
        g_main_loop_quit(mloop);
}

static gboolean
gstdin_cb(GIOChannel *gioc, GIOCondition condition, gpointer data)
{
    XfMusic *xfmusic = data;
    GList **l = NULL;
    gchar buf[3], c;
    gint i, bin = 0;
    GMainLoop *mloop = NULL;
    XfMusicState state;
    gboolean is_control_seq = FALSE;
    
    if(G_IO_STATUS_NORMAL == g_io_channel_read_chars(gioc, buf, 3, &bin, NULL)) {
        for(i=0; i < bin; i++) {
            c = buf[i];
            
            switch(c) {
                case 27:
                    is_control_seq = TRUE;
                    break;
                
                case 'A':  /* up arrow */
                case 'x':  /* kp_up */
                    if(is_control_seq)
                        xfmusic_seek(xfmusic, xfmusic_get_position(xfmusic)+30);
                    break;
                
                case 'B':  /* down arrow */
                case 'r':  /* kp_down */
                    if(is_control_seq)
                        xfmusic_seek(xfmusic, xfmusic_get_position(xfmusic)-30);
                    break;
                
                case 'C':  /* right arrow */
                case 'v':  /* kp_right */
                    if(is_control_seq)
                        xfmusic_seek(xfmusic, xfmusic_get_position(xfmusic)+5);
                    break;
                
                case 'D':  /* left arrow */
                case 't':  /* kp_left */
                    if(is_control_seq)
                        xfmusic_seek(xfmusic, xfmusic_get_position(xfmusic)-5);
                    break;
                
                case '5':  /* pgup */
                case 'y':  /* kp_pgup */
                    if(is_control_seq)
                        xfmusic_seek(xfmusic, xfmusic_get_position(xfmusic)+60);
                    break;
                
                case '6':  /* pgdown */
                case 's':  /* kp_pgdown */
                    if(is_control_seq)
                        xfmusic_seek(xfmusic, xfmusic_get_position(xfmusic)-60);
                    break;
                                
                case 'q':
                    mloop = g_object_get_data(G_OBJECT(xfmusic),
                            "xfm-cli-mainloop");
                    l = g_object_get_data(G_OBJECT(xfmusic),
                            "xfm-cli-listpointer");
                    *l = g_list_last(*l);
                    xfmusic_stop(xfmusic);
                    stream_ended_cb(xfmusic, mloop);
                    return FALSE;
                
                case 'f':
                    mloop = g_object_get_data(G_OBJECT(xfmusic),
                            "xfm-cli-mainloop");
                    xfmusic_stop(xfmusic);
                    stream_ended_cb(xfmusic, mloop);
                    break;
                
                case 'b':
                    mloop = g_object_get_data(G_OBJECT(xfmusic),
                            "xfm-cli-mainloop");
                    l = g_object_get_data(G_OBJECT(xfmusic),
                            "xfm-cli-listpointer");
                    
                    if(!(*l)->prev)
                        xfmusic_seek(xfmusic, 0);
                    else {
                        if(xfmusic_get_position(xfmusic) < 4) {
                            *l = (*l)->prev;
                            if((*l)->prev)
                                *l = (*l)->prev;
                        } else
                            *l = (*l)->prev;
                    }
                    xfmusic_stop(xfmusic);
                    stream_ended_cb(xfmusic, mloop);
                    break;
                
                case 'p':
                    state = xfmusic_get_state(xfmusic);
                    if(state == XFMUSIC_PAUSED)
                        xfmusic_play(xfmusic);
                    else if(state == XFMUSIC_PLAYING)
                        xfmusic_pause(xfmusic);
                    break;
            }
        }
    }
    
    return TRUE;
}    

static gboolean
pos_cb(gpointer user_data)
{
    XfMusic *xfmusic = user_data;
    XfMusicStreamInfo *sinfo;
    guint pos;
    static gint bitrate_avg = 0;
    static gint bitrate_nsamples = 0;
    static XfMusicStreamInfo *old_sinfo = NULL;
    
    if(xfmusic_get_state(xfmusic) != XFMUSIC_CLOSED) {
        pos = xfmusic_get_position(xfmusic);
        sinfo = xfmusic_get_stream_info(xfmusic);
        if(sinfo != old_sinfo) {
            bitrate_avg = bitrate_nsamples = 0;
            old_sinfo = sinfo;
        }
        g_print("\r [ %02d:%02d:%02d ", pos/3600, (pos/60)%60, pos%60);
        if(sinfo) {
            if(sinfo->length > 0) {
                g_print("/ %02d:%02d:%02d ]", sinfo->length/3600,
                        (sinfo->length/60)%60, sinfo->length%60);    
            } else
                g_print("]           ");
            if(sinfo->is_vbr) {
                bitrate_nsamples++;
                bitrate_avg = bitrate_avg * (((gdouble)bitrate_nsamples-1)/bitrate_nsamples)
                        + sinfo->bitrate_instant * (1.0/bitrate_nsamples);
            } else
                bitrate_avg = sinfo->bitrate_instant;
            g_print(" %s %4d kbps, %5d Hz, %s\r", sinfo->is_vbr?"ABR":"   ",
                    round((gdouble)bitrate_avg / 1000), sinfo->samples_per_sec,
                    sinfo->channels==1 ? _("mono  ") : _("stereo"));
        } else
            g_print("]\r");
    }
    
    return TRUE;
}

int
main(int argc, char **argv)
{
    GMainLoop *mloop = NULL;
    XfMusic *xfmusic;
    gint c, i, pos_cb_id, tty_fd, gio_id, long_index;
    GList *l, *files = NULL, *dummy = NULL;
    GError *err = NULL;
    gint verbose_level = 1;
    struct termios tty, tty_save;
    GIOChannel *gstdin;
    gchar *input_type = NULL, *output_type = NULL, *output_file = NULL, *p;
    
    g_thread_init(NULL);
    g_type_init();
    p = g_strrstr(argv[0], G_DIR_SEPARATOR_S);
    if(!p)
        p = argv[0] -1;
    g_set_prgname(p+1);
    
    opterr = 0;
    while((c = getopt_long(argc, argv, shortopts, longopts, &long_index)) != EOF) {
        switch(c) {
            case 'v':
                verbose_level = 2;
                break;
            
            case 'q':
                verbose_level = 0;
                break;
            
            case 'h':
                usage();
                return 0;
            
            case 'V':
                version();
                return 0;
            
            case 'i':
                if(optarg)
                    input_type = optarg;
                break;
            
            case 'o':
                if(optarg) {
                    p = strstr(optarg, ":");
                    if(p) {
                        output_type = g_strndup(optarg, p-optarg);
                        output_file = g_strdup(p+1);
                    } else
                        output_type = g_strdup(optarg);
                }
                break;
            
            default:
                g_printerr(_("Unknown option or missing argument: '%s'\n\n"), argv[optind-1]);
                usage();
                return 1;
        }
    }
    
    for(i = optind; i < argc; i++)
        files = g_list_append(files, argv[i]);
    
    if(!files) {
        usage();
        return 1;
    }
    
    mloop = g_main_loop_new(NULL, FALSE);
    g_main_loop_ref(mloop);
    
    tty_fd = open("/dev/tty", O_RDONLY);
    tcgetattr(tty_fd, &tty_save);
    tty = tty_save;
    tty.c_lflag &= ~(ECHO | ICANON);
    tty.c_cc[VMIN] = tty.c_cc[VTIME] = 0;
    tcsetattr(tty_fd, TCSAFLUSH, &tty);
    
    xfmusic = xfmusic_new(output_type);
    g_object_set_data(G_OBJECT(xfmusic), "xfm-cli-listpointer", &l);
    g_object_set_data(G_OBJECT(xfmusic), "xfm-cli-mainloop", mloop);
    g_signal_connect(G_OBJECT(xfmusic), "stream-ended",
            G_CALLBACK(stream_ended_cb), mloop);
    if(verbose_level != 0)
        pos_cb_id = g_timeout_add(500, pos_cb, xfmusic);
    gstdin = g_io_channel_unix_new(tty_fd);
    gio_id = g_io_add_watch(gstdin, G_IO_IN|G_IO_PRI, gstdin_cb, xfmusic);
    
    dummy = g_list_prepend(files, files->data);  /* for navigation */
    for(l = files; l; l = l->next) {
        if(xfmusic_open_stream(xfmusic, (const gchar *)l->data, input_type, &err)) {
            if(verbose_level == 2)
                print_stream_info(xfmusic);
            xfmusic_play(xfmusic);
            g_main_loop_run(mloop);
            xfmusic_close_stream(xfmusic);
            if(verbose_level != 0)
                g_print("\n");
        } else {
            if(err) {
                g_printerr(_("Unable to play file '%s':\n\t%s\n"),
                        (gchar *)l->data, err->message);
                g_error_free(err);
                err = NULL;
            }
        }
    }
    
    g_source_remove(gio_id);
    g_source_remove(pos_cb_id);
    g_object_unref(G_OBJECT(xfmusic));
    
    tcsetattr(tty_fd, TCSAFLUSH, &tty_save);
    close(tty_fd);
    
    g_main_loop_unref(mloop);
    g_list_free(dummy);
    
    if(output_type)
        g_free(output_type);
    if(output_file)
        g_free(output_file);
        
    return 0;
}
